start molecule
echo

Memory heap 50 stack 1500 global 500 mb

charge {{charge|default(0)}}

geometry noautoz noautosym units angstrom
  {{xyz_structure}}
end

basis spherical
 * library {{basis|default('Def2-TZVP')}}
end

{% if theory|lower == 'rhf' or theory|lower == 'scf' %}
{% set task = 'scf' %}
{% elif theory|lower == 'dft' %}
{% set task = 'dft' %}
dft
  xc {{functional|default('b3lyp')}}
  mult {{multiplicity|default(1)}}
end
{% elif theory|lower == 'mp2' %}
{% set task = 'mp2' %}
mp2
  # Exclude core electrons from MP2 treatment:"
  freeze atomic
end
{% elif theory|lower == 'ccsd' %}
{% set task = 'ccsd' %}
ccsd
  # Exclude core electrons from coupled cluster perturbations:"
  freeze atomic
end
{% endif %}

set movecs:tanalyze 0.001

{% if energy is defined  %}
task {{task}} energy
{% endif %}
{% if optimization is defined  %}
task {{task}} optimize
{% endif %}
{% if vibrational is defined  %}
task {{task}} freq
{% endif %}
